import easyocr
from dataclasses import dataclass
from lxml import etree
from rtree import index

from tocky.ia import get_page_image
from tocky.utils import Rect

# this needs to run only once to load the model into memory
reader = easyocr.Reader(['en'])

def y_overlap(rect1: Rect, rect2: Rect) -> int:
    """
    <WORD coords="169,1633,418,1568" x-confidence="99.19">Susceptance</WORD>
    near: <WORD coords="64,1618,140,1514" x-confidence="99.03">1-6</WORD>
    (1568, 1633) -- (1514, 1618)
            (1568, 1618)

             [1568           1633]
    [1514           1618]
    """
    intersection_top = max(rect1.top, rect2.top)
    intersection_bottom = min(rect1.bottom, rect2.bottom)

    return max(0, intersection_bottom - intersection_top)

@dataclass
class OCRWord:
  index: int
  text: str
  rect: Rect
  conf: float  # Number between 0 and 1

  def to_xml(self):
    # <WORD coords="241,509,757,421" x-confidence="96">Psychological</WORD>
    el = etree.Element('WORD')
    el.set('coords', f'{round(self.rect.left)},{round(self.rect.bottom)},{round(self.rect.right)},{round(self.rect.top)}')
    el.set('x-confidence', f'{(self.conf * 100):.2f}')
    el.text = self.text
    return el

def index_words(words: list[OCRWord]) -> index.Index:
  rect_idx = index.Index()

  for word in words:
    rect_idx.insert(word.index, word.rect.to_ltrb())

  return rect_idx


def group_into_lines(words: list[OCRWord], max_width: int) -> list[list[OCRWord]]:
  rect_idx = index_words(words)

  word_i_to_line_i = {}
  max_line_id = 0
  for word in words:
    # print(word.text)
    mid_y = word.rect.top + (word.rect.bottom - word.rect.top) / 2
    for nearby_word_i in rect_idx.nearest((0, mid_y, max_width, mid_y), 5):
      if nearby_word_i == word.index:
        continue
      nearby_word = words[nearby_word_i]
      overlap = y_overlap(word.rect, nearby_word.rect) / word.rect.height
      # print(f'    near "{nearby_word.text}" ({overlap=})')
      if overlap > 0.35 and nearby_word.index in word_i_to_line_i:
        # print(repr(word.text), '->', word_i_to_line_i[nearby_word.index])
        word_i_to_line_i[word.index] = word_i_to_line_i[nearby_word.index]
        break
    else:
      # print('    ', repr(word.text), '->', max_line_id)
      max_line_id += 1
      word_i_to_line_i[word.index] = max_line_id

  lines = [[] for _ in range(max_line_id+1)]
  for word_id, line_id in word_i_to_line_i.items():
    lines[line_id].append(words[word_id])
  for line in lines:
    line.sort(key=lambda word: (word.rect.left // 50, word.rect.top))
  return lines


def ocr_djvu_page_easyocr(ocaid: str, leaf_num: int) -> str:
  print('ocr_djvu_page_easyocr', ocaid, leaf_num)
  img = get_page_image(ocaid, leaf_num)
  # return reader.readtext(img)
  words = [
      OCRWord(index=i, text=text, rect=Rect.from_cw_points(cw_points), conf=conf)
      for i, (cw_points, text, conf) in enumerate(
          (cw_points, text, conf)
          for (cw_points, text, conf) in reader.readtext(img)
          if text and not (len(text) <= 2 and conf < 0.1)
      )
  ]

  par_el = etree.Element('PARAGRAPH')
  for line in group_into_lines(words, img.width):
    line_el = etree.Element('LINE')
    for word in line:
      line_el.append(word.to_xml())
    par_el.append(line_el)
  return '<HIDDENTEXT x-re-ocrd="true"><PAGECOLUMN><REGION>' + etree.tostring(par_el, encoding='utf-8').decode('utf-8') + '</REGION></PAGECOLUMN></HIDDENTEXT>'



import pytesseract
from PIL import Image
from io import StringIO
from lxml import etree
import pandas as pd
import csv

from tocky.ia import get_ia_metadata, get_page_image
from tocky.ocr.tesseract.preprocessing import remove_leader_dots

def ocr_image(img: Image, psm: int, dpi: int):
  # --psm 11: Sparse text. Find as much text as possible in no particular order.
  ocr_tsv_data = pytesseract.image_to_data(remove_leader_dots(img), config=f'--psm {psm} --dpi {dpi} -l eng')
  df = pd.read_csv(StringIO(ocr_tsv_data), sep='\t', quoting=csv.QUOTE_NONE)
  df = df[
      (~df.text.isna())
      & (~df['text'].isin(['', *' |{}\\_=—']))
      & (df['width'] > 0)
  ]
  # Round to make displaying lines/etc more easily aligned
  # df['left'] = df['left'].apply(lambda n: 10 * (n // 10))
  # df['top'] = df['top'].apply(lambda n: 10 * (n // 10))
  return df.reset_index()

def ocr_djvu_page_tesseract(ocaid: str, leaf_num: int) -> str:
  """
  OCR a page of a djvu using tesseract and return the result as a HIDDENTEXT DjVu XML string
  """
  img = get_page_image(ocaid, leaf_num)
  df = ocr_image(img, psm=3, dpi=int(get_ia_metadata(ocaid)['ppi']))


  ## Find the closest rect in the --psm 4 table that possible contains the rect
  # low_conf_words = df[df['conf'] < 90]
  # ocr_psm_4 = ocr_image(img, psm=4)
  # for index, lc_row in low_conf_words.iterrows():
  #   ocr_psm_4['distance'] = ocr_psm_4.apply(lambda row: abs(row.left - lc_row.left) + abs(row.top - lc_row.top), axis=1)
  #   best = ocr_psm_4.sort_values('distance').iloc[0]
  #   if best.distance < 5 and best.text != lc_row.text and best.conf > lc_row.conf:
  #     #print(index, (lc_row.text, lc_row.conf), '→', (best.text, best.conf), best.distance)

  #     df.loc[index, 'text'] = best.text
  #     df.loc[index, 'conf'] = best.conf
  #     df.loc[index, 'left'] = best.left
  #     df.loc[index, 'top'] = best.top
  #     df.loc[index, 'width'] = best.width
  #     df.loc[index, 'height'] = best.height

  lines = []
  for _, group in df.groupby(['block_num', 'par_num', 'line_num']):
    line_el = etree.Element('LINE')
    for index, row in group.iterrows():
      # <WORD coords="241,509,757,421" x-confidence="96">Psychological</WORD>
      word_el = etree.Element('WORD')
      word_el.set('coords', f'{row.left},{row.top + row.height},{row.left + row.width},{row.top}')
      word_el.set('x-confidence', str(row.conf))
      word_el.text = row.text
      line_el.append(word_el)
    lines.append(etree.tostring(line_el, encoding='utf-8').decode('utf-8'))
  return '<HIDDENTEXT x-re-ocrd="true"><PAGECOLUMN><REGION><PARAGRAPH>' + '\n'.join(lines) + '</PARAGRAPH></REGION></PAGECOLUMN></HIDDENTEXT>'



from dataclasses import dataclass, field
import re
import tiktoken
from typing import Literal
import json
import traceback
import os
import requests

from tocky.detector import extract_toc_pages
from tocky.extractor import clean_raw_toc_ocr
from tocky.extractor.printer import extract_djvu_page_text
from tocky.ia import extract_page_index, get_ia_metadata
from tocky.validator import validate_extracted_toc

@dataclass
class ItemProcessingState:
  ocaid: str
  status: Literal['Already has TOC', ''] = ''
  toc_raw_ocr: str = None
  detected_toc: list[str] = field(default_factory=list)
  toc_ocr: str = ''
  structured_toc: str = ''
  error: Exception = None
  prompt_tokens: int = 0
  completion_tokens: int = 0


def process_ol_book(ol_record: dict) -> ItemProcessingState:
  state = ItemProcessingState(ocaid=ol_record['ocaid'])
  ol_toc = ol_record.get('table_of_contents')
  if ol_toc:
    toc_missing_pagenums = not any(chapter.get('pagenum') for chapter in ol_toc)
    if ol_toc and not toc_missing_pagenums:
      state.status = 'Already has good TOC'
      return state

  try:
    state.detected_toc = list(extract_toc_pages(state.ocaid))
  except Exception as e:
    state.status = 'Errored'
    state.error = e
    return state

  if not state.detected_toc:
    state.status = 'No TOC detected'
  else:
    try:
      state.toc_raw_ocr = [
          extract_djvu_page_text(djvu_xml, re_ocr=True)
          for (page, djvu_xml) in state.detected_toc
      ]
    except Exception as e:
      state.status = 'Errored'
      state.error = e

    if re.search(r'([A-Za-z]{25,}|\beee+\b)', '\n'.join(state.toc_raw_ocr), re.MULTILINE):
      state.status = 'Bad OCR on TOC'
    else:
      chunks = ['']
      for page_ocr in state.toc_raw_ocr:
        extended_chunk = chunks[-1] + '\n' + page_ocr
        if len(tiktoken.encoding_for_model("gpt-3.5-turbo").encode(extended_chunk)) > 1000:
          chunks.append(page_ocr)
        else:
          chunks[-1] += '\n' + page_ocr
      try:
        book_title = get_ia_metadata(state.ocaid)['title']
        for chunk in chunks:
          toc_response = clean_raw_toc_ocr(chunk, book_title, prev_toc=state.structured_toc)
          state.structured_toc += '\n' + toc_response.toc.strip().strip('`').strip()
          state.prompt_tokens += toc_response.prompt_tokens
          state.completion_tokens += toc_response.completion_tokens
        state.status = 'TOC Extracted'
      except Exception as e:
        state.status = 'Errored'
        state.error = e
      if state.structured_toc:
        total_pages = int(get_ia_metadata(state.ocaid)['imagecount'])
        validation = validate_extracted_toc(state.structured_toc, total_pages)
        if validation != 'Valid':
          state.status = f'TOC Validation: {validation}'

  return state

def push_to_toc_queue(record: dict):
  return requests.put(
      'https://testing.openlibrary.org/tocky/push',
      headers={
          'X-API-KEY': os.environ['TOC_QUEUE_DB_PASSWORD'],
          'Content-Type': 'application/json',
      },
      data=json.dumps(record)
  )

def process_all(ia_query: str, rows=10, page=1, ia_overrides=None):
  ia_overrides = ia_overrides or {}
  ia_records = requests.get('https://archive.org/advancedsearch.php', params={
    **ia_query,
    'fl': 'identifier,openlibrary_edition',
    'rows': rows,
    'page': page,
    'output': 'json',
  }).json()['response']['docs']

  for ia_record in ia_records:
    if ia_record['identifier'] in ia_overrides:
      ia_record |= ia_overrides[ia_record['identifier']]

  ol_records_by_key = requests.get('http://openlibrary.org/api/get_many', params={
      'keys': json.dumps([
          f'/books/{metadata["openlibrary_edition"]}'
          for metadata in ia_records
      ])
  }).json()['result']

  for ia_record in ia_records:
    ol_record = ol_records_by_key[f'/books/{ia_record["openlibrary_edition"]}']
    if 'ocaid' not in ol_record:
      ol_record['ocaid'] = ia_record['identifier']

  import concurrent.futures
  all_results = []
  with concurrent.futures.ThreadPoolExecutor(max_workers=5) as executor:
    for result in executor.map(process_ol_book, ol_records_by_key.values()):
      print(f'[{result.status}] {result.ocaid}')
      push_to_toc_queue({
          'code_version': 'v2.E.1',
          'ocaid': result.ocaid,
          'status': result.status,
          'prompt_tokens': result.prompt_tokens,
          'completion_tokens': result.completion_tokens,
          'error': str(result.error) if result.error else None,
          'toc_raw_ocr': result.toc_raw_ocr,
          'structured_toc': result.structured_toc,
          'detected_toc': [extract_page_index(p[0]) for p in result.detected_toc],
      })
      if result.error:
        print(traceback.print_exception(result.error))
      all_results.append(result)
  return all_results



import re
from dataclasses import dataclass
import openai

@dataclass
class TocResponse:
  toc: str
  prompt_tokens: int
  completion_tokens: int

def clean_raw_toc_ocr(raw_ocr: str, book_title: str, prev_toc: str = None) -> TocResponse:
  prompt = """
You are a librarian extracting table of contents data in a structured format. The format you will need to output is as follows:

```
* {label (optional)} | {title} | {page number}
```

### Important Instructions:
- Fix any typos or bad characters generated by the OCR.
- Change any text in ALL UPPERCASE to be in Normal Title Case.

### Examples:

Input:
```
                                   PAGE
PREFACE                             1x
PART 1: THIS VVORLD                  1
   Chapter I.
   Of the Nature of Flatland
                                      3
   Chapter IT.
   Of the Cl'mate and Houses in
       Hlatland  . . . dsa. .a;;; ... 5
PART 2: OTHER WORLDS
                                    42
```

Output:
```
* | Preface | ix
* Part 1 | This World | 1
    ** Chapter I | Of the Nature of Flatland | 3
    ** Chapter II | Of the Climate and Houses in Flatland | 5
* Part 2 | Other Worlds | 42
```


Input:
```
      CONTENTS
CHapter 1. ...... ... . . 1
Ch^pter 2            .. @ 25
Chaptr               .   38
Chaptr  4                 48
```
Output
```
* | Chapter 1 | 1
* | Chapter 2 | 25
* | Chapter 3 | 38
* | Chapter 4 | 48
```

You can nest when necessary:

Input:
```
A. Technology
     Computers     ...       1
          Hard-drives        2
          Software           8
    II. Machinery           1[
   III. Hardware            3/
B. Agriculture
```

```
* A | Technology |
    ** I | Computers | 1
        *** | Hard-drives | 2
        *** | Software | 8
    ** II | Machinery | 11
    ** III | Hardware | 37
* B | Agriculture |
```
"""

  if prev_toc:
    nline = '\n'
    message = f"""
The last table of contents was too long too fit in your context. Continue extracting the table of contents of "{book_title}".

Here is the end of what you last output. Do not output this again, though.

```
{nline.join(prev_toc.split(nline)[-10:])}
```

And here is the rest of the OCR text:
```
{raw_ocr}
```
""".strip()
  else:
    message = f"""
Extract the table of contents from this OCR text of "{book_title}":

```
{raw_ocr}
```
  """.strip()
  print('openai request', message[0:500] + '...')
  completion = openai.chat.completions.create(
    model="gpt-3.5-turbo-1106",
    messages=[
      {"role": "system", "content": prompt},
      {"role": "user", "content": message},
    ],
    # max_tokens=1024,
    n=1,
    stop=None,
    temperature=0.5,
  )

  def clean_line(line: str) -> str:
    pipes = len(line.split('|')) - 1
    m = re.search(r'^ *\*+', line)
    if not m:
      return line
    stars = m.group()
    rest = line[m.span()[1]:]
    if pipes == 0:
      return f'{stars} | {rest.strip()} |'
    elif pipes == 1:
      return f'{stars} | {rest.strip()}'
    elif pipes == 2:
      return line
    elif pipes == 3:
      return line.replace('|', ' ', 1)
    else:
      # Not good
      return line

  toc = completion.choices[0].message.content
  prompt_tokens = completion.usage.prompt_tokens
  completion_tokens = completion.usage.completion_tokens
  return TocResponse(
      '\n'.join(map(clean_line, toc.split('\n'))),
      prompt_tokens,
      completion_tokens)


import functools
from io import BytesIO
import itertools
import re
import sys
from internetarchive import get_session
from PIL import Image
from lxml import etree

ia_session = get_session()

@functools.cache
def get_ia_metadata(ocaid: str) -> dict:
  return ia_session.get(f"https://archive.org/metadata/{ocaid}/metadata").json()['result']

def get_page_image(ocaid: str, leaf_num: int) -> Image:
  img = ia_session.get(f"https://archive.org/download/{ocaid}/{ocaid}_jp2.zip/{ocaid}_jp2%2F{ocaid}_{leaf_num:04}.jp2", params={
      'ext': 'jpg',
      'reduce': '0',
      'quality': '100',
  })

  return Image.open(BytesIO(img.content))

def extract_page_index(page_filename: str) -> int:
  return int(re.search(r'(?:_)(\d+)(?:\.djvu)', page_filename).group(1))

def ocaid_to_djvu_url(ocaid: str) -> str:
  get_ia_metadata(ocaid)
  return f"https://archive.org/download/{ocaid}/{ocaid}_djvu.xml"

def get_djvu_pages(djvu_url: str, start: int=0, end: int=sys.maxsize):
    response = ia_session.get(djvu_url, stream=True)

    if response.status_code != 200:
        print("Error: Unable to fetch the DJVU file.")
        return

    response.raw.decode_content = True
    for _, elem in itertools.islice(etree.iterparse(response.raw, events=("end",), tag="OBJECT"), start, end):
      page_name = elem.xpath(".//PARAM[@name='PAGE']/@value")[0]
      yield page_name, elem


import math
from lxml import etree
from rtree import index

from tocky.ia import extract_page_index, get_ia_metadata
from tocky.ocr import ocr_djvu_page
from tocky.utils import avg_ocr_conf

def extract_djvu_page_text_old(djvu_page: str | etree._Element) -> str:
    if isinstance(djvu_page, str):
       root = etree.fromstring(djvu_page)
    else:
        root = djvu_page
    page_file = root.xpath('./@usemap')[0]

    result = f"\n\n----------- OCR of page {extract_page_index(page_file)} -----------\n\n\n"

    ref_word = next(
        word
        for word in root.findall('.//WORD')
        if len(word.text) > 8
    )
    ref_word_coords = ref_word.xpath("./@coords")[0].split(',')
    ref_word_width = int(ref_word_coords[2]) - int(ref_word_coords[0])
    char_width = ref_word_width / len(ref_word.text)

    for pagecol in root.findall('.//PAGECOLUMN'):
        for par in pagecol.findall('.//PARAGRAPH'):
            for line in par.findall('.//LINE'):
                words = line.findall('.//WORD')
                if words:
                    coords = line.xpath('.//WORD/@coords')[0].split(',')

                    line_indent = math.floor(int(coords[0]) / char_width)
                    line_text = ' '.join(word.text.strip() for word in words)
                    indented_line = ' ' * line_indent + line_text
                    result += indented_line + '\n'

    return result

def extract_djvu_page_text(djvu_page: str | etree._Element, re_ocr=False) -> str:
    if isinstance(djvu_page, str):
       root = etree.fromstring(djvu_page)
    else:
        root = djvu_page
    canvas = ""
    page_file = root.xpath('./@usemap')[0]
    ocaid = page_file.rsplit('_', 1)[0]
    leaf_num = extract_page_index(page_file)

    # Convert inches to pixels
    dpi = int((root.xpath(".//PARAM[@name='DPI']/@value") or [int(get_ia_metadata(ocaid)['ppi'])])[0])
    line_rounding_size = round(0.0333 * dpi)

    if re_ocr and root.xpath('.//HIDDENTEXT/@x-re-ocrd') != ['true']:
      new_ocr = ocr_djvu_page(ocaid, leaf_num)
      new_ocr_el = etree.fromstring(new_ocr)
      if (avg_ocr_conf(new_ocr_el) or 100) > (avg_ocr_conf(root.find('.//HIDDENTEXT')) or 0):
        root.replace(root.find('.//HIDDENTEXT'), new_ocr_el)

    canvas_width = 120  # dpi // 4
    conv_factor = canvas_width / int(root.xpath('./@width')[0])
    canvas_height = math.ceil(int(root.xpath('./@height')[0]) * conv_factor)
    canvas = ((" " * canvas_width + '\n') * canvas_height)[:-1]

    # if Path('./tmp.log').exists():
    #   Path('./tmp.log').unlink()

    # def log(message: str) -> None:
    #   log_file = Path('./tmp.log')

    #   # Ensure the file and its parent directory exist
    #   log_file.parent.mkdir(parents=True, exist_ok=True)

    #   # Append the message to the file
    #   with open(log_file, 'a') as file:
    #       file.write(message + '\n')

    # log(f'{conv_factor=} {canvas_width=} {canvas_height=}')
    used_regions = index.Index()
    rect_index = 0
    for pagecol in root.findall('.//PAGECOLUMN'):
        for par in pagecol.findall('.//PARAGRAPH'):
            for line in par.findall('.//LINE'):
                # log('LINE')
                line_row = None
                for word in line.findall('.//WORD'):
                    coords = word.xpath('./@coords')[0].split(',')
                    left = int(coords[0])
                    bottom = int(coords[1])
                    right = int(coords[2])
                    top = int(coords[3])

                    baseline = round((top + (bottom - top) / 3) / line_rounding_size) * line_rounding_size
                    line_row = math.floor(baseline * conv_factor) if line_row is None else line_row
                    col = math.floor(left * conv_factor)

                    while list(used_regions.intersection(rect := (col, line_row, col + len(word.text), line_row + 1))):
                      col += 1

                    used_regions.insert(rect_index, rect)
                    rect_index += 1

                    start_idx = (canvas_width + 1) * line_row + col
                    end_idx = start_idx + len(word.text)
                    shift = 0
                    max_shift = 5
                    # log(f'{line_row}:{col}\t{coords}\t{word.text}')
                    # while not canvas[max(0, start_idx - 1):(end_idx + 1)].isspace():
                    #   start_idx += 1
                    #   end_idx += 1
                    #   shift += 1
                    #   if shift >= max_shift:
                    #     break
                    canvas = canvas[0:start_idx] + word.text + canvas[end_idx:]

    def crop_text_canvas(canvas: str) -> str:
      lines = canvas.split('\n')
      line_len = len(lines[0])
      top = 0
      left = line_len
      right = 0
      bottom = 0

      for line in lines:
        left = min(left, line_len - len(line.lstrip()))
        right = max(right, len(line.rstrip()))

      for line in lines:
        if not line.isspace():
          break
        top += 1

      for line in reversed(lines):
        if not line.isspace():
          break
        bottom += 1
      return '\n'.join((
          line[left:right] if not line.isspace() else ''
          for line in lines[top:-max(bottom, 1)]
      ))

    def collapse_newlines(s: str):
        import re
        min_newline_chain = min(map(len, re.findall(r'\n+', s.strip(), re.MULTILINE)), default=1)

        return re.sub(r'\n' * (min_newline_chain -1) + '(\n*)', r'\1', s, flags=re.MULTILINE)
        # return re.sub(r'\n+', '\n', s, flags=re.MULTILINE)

    def collapse_spaces(s: str):
        import re
        min_newline_chain = min(map(len, re.findall(r' +', s.strip(), re.MULTILINE)), default=1)

        return re.sub(r' ' * (min_newline_chain -1) + '( *)', r'\1', s, flags=re.MULTILINE)
        # return re.sub(r'(\S) +', r'\1 ', s)

    return collapse_newlines(crop_text_canvas(canvas))



def ocr_djvu_page(ocaid: str, leaf_num: int, engine='easyocr') -> str:
  if engine == 'easyocr':
    from tocky.ocr.easyocr import ocr_djvu_page_easyocr
    return ocr_djvu_page_easyocr(ocaid, leaf_num)
  elif engine == 'tesseract':
    from tocky.ocr.tesseract import ocr_djvu_page_tesseract
    return ocr_djvu_page_tesseract(ocaid, leaf_num)
  else:
    raise ValueError(f'Unknown OCR engine {engine}')


from dataclasses import dataclass
import re
from lxml import etree

@dataclass
class Rect:
  x: int
  y: int
  width: int
  height: int
  left: int
  right: int
  bottom: int
  top: int

  @staticmethod
  def from_xywh(rect: tuple) -> 'Rect':
    x, y, w, h = rect
    return Rect(
        x=x,
        y=y,
        width=w,
        height=h,
        left=x,
        right=x + w,
        bottom=y + h,
        top=y,
    )

  @staticmethod
  def from_ltrb(rect: tuple) -> 'Rect':
    l, t, r, b = rect
    return Rect(
        x=l,
        y=t,
        width=r - l,
        height=b - t,
        left=l,
        right=r,
        bottom=b,
        top=t,
    )

  @staticmethod
  def from_cw_points(points: list[list[int]]) -> 'Rect':
    # Points can be rotated
    l = min(p[0] for p in points)
    t = min(p[1] for p in points)
    r = max(p[0] for p in points)
    b = max(p[1] for p in points)
    return Rect.from_ltrb((l, t, r, b))

  def to_ltrb(self) -> tuple:
    return (self.left, self.top, self.right, self.bottom)

def pretty_print_xml(root: etree._Element) -> str:
  import xml.dom.minidom
  result = xml.dom.minidom.parseString(etree.tostring(root)).toprettyxml(indent='  ')
  return strip_newlines(result)

def strip_newlines(text: str) -> str:
  result = re.sub(r'^\s+$', '', text, flags=re.MULTILINE)
  result = re.sub(r'\n+', '\n', result)
  return result

def avg_ocr_conf(el: etree._Element) -> float:
  conf_sum = 0
  conf_count = 0
  for word in el.findall('.//WORD'):
    if word.xpath('./@x-confidence'):
      conf_sum += float(word.xpath('./@x-confidence')[0])
      conf_count += 1
  if conf_count:
    return conf_sum / conf_count


from dataclasses import dataclass
import itertools
import re

from lxml import etree

from tocky.ia import extract_page_index, get_djvu_pages, ocaid_to_djvu_url
from tocky.ocr import ocr_djvu_page
from tocky.utils import avg_ocr_conf

TOC_PAGE_DETECTOR_VERSION = [
    ('v2.E.1', 'Fix: Don\'t check every page after TOC found!'),
    ('v2.E.0', 'Add support for re-ocring when ending TOC range to avoid missing last page(s)'),
    ('v2.D.0', 'Add support for re-ocring previous page once first TOC page detected to avoid missing first page'),
]

@dataclass
class TOCDetectorHyperParams:
  min_lines_with_nums: int = 5
  max_lines_with_nums: int = 50
  max_repeated_nums: int = 4
  min_word_confidence: float = 40
  min_mean_word_confidence: float = 30
  min_small_nums_count_for_first: int = 4
  min_avg_word_len: float = 3.8
  min_increasing_percent: float = 0.75


@dataclass
class TOCAnalysisResult:
  is_toc: bool = False
  failure: str | None = None
  reran_ocr: bool = False
  used_new_ocr: bool = False

def analyze_page_for_toc(
  elem: etree._Element,
  has_begun = False,
  allow_reocr=True,
  redo_ocr=False,
) -> TOCAnalysisResult:
  hiddentext = elem.find(".//HIDDENTEXT")
  if hiddentext is None:
    return TOCAnalysisResult(False, 'Missing hiddentext')

  P = TOCDetectorHyperParams()
  result = TOCAnalysisResult()

  page_file: str = elem.xpath('./@usemap')[0]
  ocaid = page_file.rsplit('_', 1)[0]
  leaf_num = extract_page_index(page_file)
  # print('analyze_page_for_toc', page_file, f'{allow_reocr=}')

  def do_redo_ocr(min_conf: float):
    if elem.xpath('.//HIDDENTEXT/@x-re-ocrd') == ['true']:
      return
    new_ocr = ocr_djvu_page(ocaid, leaf_num)
    result.reran_ocr = True
    new_ocr_el = etree.fromstring(new_ocr)
    # Re-OCR will always output confidences, so assume 100 (likely empty page)
    new_conf = avg_ocr_conf(new_ocr_el) or 100
    # print(f'{page_file} mean_conf {(mean_conf or 0):.2f} -> {new_conf:.2f}')
    if new_conf > min_conf:
      result.used_new_ocr = True
      elem.replace(elem.find('.//HIDDENTEXT'), new_ocr_el)
      return analyze_page_for_toc(elem, has_begun, allow_reocr=False, redo_ocr=False)

  mean_conf = avg_ocr_conf(elem)

  if redo_ocr:
    if new_result := do_redo_ocr(min_conf=(mean_conf or 0) - 15):
      return new_result

  if mean_conf is not None and mean_conf < P.min_mean_word_confidence:
    if allow_reocr:
      if new_result := do_redo_ocr(min_conf=mean_conf):
        return new_result
    return TOCAnalysisResult(False, f'Mean page confidence too low ({mean_conf:.2f} < {P.min_mean_word_confidence})')

  # If there are more than 40 words and the average right coord of each word is
  # less than 2/3 width of page, then potentially should re-ocr cause mis-OCR'd
  # pages
  if allow_reocr:
    word_rights = [
        int(coord.split(',')[2])
        for coord in hiddentext.xpath('.//WORD/@coords')
    ]
    if len(word_rights) > 70:
      width = int(elem.xpath('./@width')[0])
      words_near_right = len([r for r in word_rights if r > width * .8 ])

      words_near_right_text = [
          word.text
          for word in hiddentext.xpath('.//WORD')
          if int(word.xpath('./@coords')[0].split(',')[2]) > width * .8
      ]
      numeric_words_near_right = len([
          word
          for word in words_near_right_text
          if re.search(r'\b([\dxvil]+)$', word) and re.search(r'\d[.,-]\d+$', word) is None
      ])
      # print(f'{words_near_right=}, {numeric_words_near_right=} of {len(word_rights)=}')
      if (words_near_right - numeric_words_near_right) < 3:
        if new_result := do_redo_ocr(min_conf=(mean_conf or 0) - 15):
          return new_result


  numbers_at_end_of_lines = []
  for line in hiddentext.findall(".//LINE"):
      words = line.findall(".//WORD")
      if words:
        last_word = words[-1]
        if (m := re.search(r'\b([\dxvil]+)$', last_word.text)) and re.search(r'\d[.,-]\d+$', last_word.text) is None:
            confidence = last_word.xpath('./@x-confidence')
            bad_ocr = confidence and float(confidence[0]) < P.min_word_confidence
            if len(m.group()) == 1 and bad_ocr:
              # Exclude, likely noise
              pass
            else:
              numbers_at_end_of_lines.append(m.group())

  from collections import Counter
  max_repeated = max(Counter(numbers_at_end_of_lines).values(), default=0)
  lines_w_uniq_nums = len(set(numbers_at_end_of_lines))

  if max_repeated > P.max_repeated_nums:
    return TOCAnalysisResult(False, 'Too many repeated numbers')

  if lines_w_uniq_nums < P.min_lines_with_nums:
    return TOCAnalysisResult(False, f'Not enough lines with nums ({lines_w_uniq_nums} < {P.min_lines_with_nums})')

  if lines_w_uniq_nums > P.max_lines_with_nums:
     return TOCAnalysisResult(False, 'Too many lines with nums')

  nums = [
      int(n.strip())
      for n in numbers_at_end_of_lines
      if n.strip().isnumeric()
  ]
  increasing_count = 0
  non_increasing_count = 0
  for (prev, cur) in itertools.pairwise([0] + nums):
    if prev <= cur:
      increasing_count += 1
    else:
      non_increasing_count += 1

  increasing_percent = increasing_count / (increasing_count + non_increasing_count)
  if len(nums) > 4 and increasing_percent < P.min_increasing_percent:
    return TOCAnalysisResult(False, f'Not enough of the nums are increasing {increasing_percent:.2f} < {P.min_increasing_percent}')

  if has_begun:
    first_line = ' '.join(word.text for word in hiddentext.find(".//LINE").findall(".//WORD"))
    if 'illustration' in first_line.lower() or 'index' in first_line.lower() or 'figure' in first_line.lower() or 'tables' in first_line.lower():
      return TOCAnalysisResult(False, 'Contains illustration/index')
  else:
    small_nums_count = len([n for n in numbers_at_end_of_lines if len(n) < 4])
    if small_nums_count < P.min_small_nums_count_for_first:
      return TOCAnalysisResult(False, 'Not enough small numbers for first TOC page')

    # Check average word length probably bad OCR on empty page
    word_lens = [
        len(word.text.strip())
        for word in elem.findall('.//WORD')
        if not re.search(r'^(\d+|[xvil]+|[\.^;:{}]+)$', word.text.strip(), re.IGNORECASE)
    ]
    avg_word_len = sum(word_lens) / len(word_lens)
    if avg_word_len < P.min_avg_word_len:
      return TOCAnalysisResult(False, f'Average word length too low {avg_word_len:.2f} < {P.min_avg_word_len}')

  if allow_reocr:
    if new_result := do_redo_ocr(min_conf=(mean_conf or 0) - 15):
      return new_result

  return TOCAnalysisResult(True)

def analyze_djvu_for_toc(djvu_url: str, index: int = None, allow_reocr=True):
    has_begun = False
    reocrs_done = 0
    last_result: tuple[str, str, TOCAnalysisResult] = None
    results: list[tuple[str, str, TOCAnalysisResult]] = []
    for page_name, elem in get_djvu_pages(djvu_url, 1, 28):
      page_index = extract_page_index(page_name)
      toc_analysis = analyze_page_for_toc(elem, has_begun, allow_reocr=allow_reocr and (has_begun or reocrs_done < 10))
      if toc_analysis.reran_ocr:
        reocrs_done += 1
      # Let's check predecessor with redone ocr
      if not has_begun and toc_analysis.is_toc and last_result and not last_result[2].reran_ocr:
        # Check if number is too high before redoing OCR
        recheck_previous = False
        for line in elem.findall(".//LINE"):
            words = line.findall(".//WORD")
            if words:
              last_word = words[-1]
              if has_roman_numeral := bool(re.search(r'\b([xvil]+)$', last_word.text)):
                recheck_previous = False
                break

              if number_m := re.search(r'\b(\d+)$', last_word.text):
                n = int(number_m[1])
                recheck_previous = n > 25
                break

        if recheck_previous:
          new_toc_analysis = analyze_page_for_toc(last_result[1], False, redo_ocr=True)
          reocrs_done += 1
          if new_toc_analysis.is_toc:
            last_result = (last_result[0], last_result[1], new_toc_analysis)

      if has_begun and not toc_analysis.is_toc and not toc_analysis.reran_ocr:
        # Let's try again with rerun ocr
        reocrs_done += 1
        toc_analysis = analyze_page_for_toc(elem, has_begun, redo_ocr=True)

      # Yield the thing we know won't change anymore
      if last_result:
        yield (last_result[0], etree.tostring(last_result[1], encoding='unicode'), last_result[2])

      has_begun = toc_analysis.is_toc
      last_result = (page_name, elem, toc_analysis)

    if last_result:
      yield (last_result[0], etree.tostring(last_result[1], encoding='unicode'), last_result[2])


def detect_table_of_contents_pages(djvu_url: str, allow_reocr=True):
  has_begun = False
  for (page_number, elem_str, toc_analysis) in analyze_djvu_for_toc(djvu_url, allow_reocr=allow_reocr):
    if has_begun and not toc_analysis.is_toc:
      break
    if toc_analysis.is_toc:
      has_begun = True
      yield (page_number, elem_str)



def extract_pages(djvu_url: str, pages: list[int]):
    to_see = set(pages)
    for page_name, elem in get_djvu_pages(djvu_url):
      page_index = extract_page_index(page_name)
      if page_index in to_see:
        yield (page_name, etree.tostring(elem, encoding="unicode"))
        to_see.remove(page_index)
        if not to_see:
          break

def extract_toc_pages(ocaid: str, allow_reocr = True):
  return detect_table_of_contents_pages(ocaid_to_djvu_url(ocaid), allow_reocr=allow_reocr)
